/* PulseAudioClip.java
   Copyright (C) 2008 Red Hat, Inc.

This file is part of IcedTea.

IcedTea is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 2.

IcedTea is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with IcedTea; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
02110-1301 USA.

Linking this library statically or dynamically with other modules is
making a combined work based on this library.  Thus, the terms and
conditions of the GNU General Public License cover the whole
combination.

As a special exception, the copyright holders of this library give you
permission to link this library with independent modules to produce an
executable, regardless of the license terms of these independent
modules, and to copy and distribute the resulting executable under
terms of your choice, provided that you also meet, for each linked
independent module, the terms and conditions of the license of that
module.  An independent module is a module which is not derived from
or based on this library.  If you modify this library, you may extend
this exception to your version of the library, but you are not
obligated to do so.  If you do not wish to do so, delete this
exception statement from your version.
 */

package org.classpath.icedtea.pulseaudio;

import java.util.concurrent.Semaphore;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.LineEvent;
import javax.sound.sampled.LineUnavailableException;

public abstract class PulseAudioDataLine extends PulseAudioLine implements
		DataLine {

	protected static final int DEFAULT_BUFFER_SIZE = StreamBufferAttributes.SANE_DEFAULT;
	protected static final String PULSEAUDIO_FORMAT_KEY = "PulseAudioFormatKey";

	protected String streamName = "Java Stream";

	// true between start() and stop()
	protected boolean isStarted = false;

	protected boolean isEngagedInIo = false;

	// true if a stream has been paused
	// protected boolean isPaused = false;

	protected AudioFormat[] supportedFormats = null;
	protected AudioFormat currentFormat = null;
	protected AudioFormat defaultFormat = null;

	protected int bufferSize = 0;

	protected EventLoop eventLoop = null;
	protected Semaphore semaphore = new Semaphore(0);
	protected Stream stream;

	public void open(AudioFormat format, int bufferSize)
			throws LineUnavailableException {

		if (isOpen) {
			throw new IllegalStateException("DataLine is already open");
		}

		for (AudioFormat myFormat : supportedFormats) {
			if (format.matches(myFormat)) {
				synchronized (eventLoop.threadLock) {
					stream = new Stream(eventLoop.getContextPointer(),
							streamName, Stream.Format.valueOf((String) myFormat
									.getProperty(PULSEAUDIO_FORMAT_KEY)),
							(int) format.getSampleRate(), format.getChannels());

				}
				currentFormat = format;
				super.open();
			}
		}
		// no matches found
		if (!isOpen) {
			throw new IllegalArgumentException("Invalid format");
		}

		Stream.StateListener openCloseListener = new Stream.StateListener() {

			@Override
			public void update() {
				synchronized (eventLoop.threadLock) {

					/*
					 * Not the order: first we notify all the listeners, and
					 * then return. this means no race conditions when the
					 * listeners are removed before they get called by close
					 * 
					 * eg:
					 * 
					 * line.close(); line.removeLineListener(listener)
					 * 
					 * the listener is guaranteed to have run
					 * 
					 */

					if (stream.getState() == Stream.State.READY) {
						fireLineEvent(new LineEvent(PulseAudioDataLine.this,
								LineEvent.Type.OPEN, AudioSystem.NOT_SPECIFIED));
						semaphore.release();
					} else if (stream.getState() == Stream.State.TERMINATED
							|| stream.getState() == Stream.State.FAILED) {
						fireLineEvent((new LineEvent(PulseAudioDataLine.this,
								LineEvent.Type.CLOSE, AudioSystem.NOT_SPECIFIED)));
						semaphore.release();
					}
				}
			}
		};

		stream.addStateListener(openCloseListener);

		Stream.UnderflowListener stoppedListener = new Stream.UnderflowListener() {
			@Override
			public void update() {
				isEngagedInIo = false;
				fireLineEvent(new LineEvent(PulseAudioDataLine.this,
						LineEvent.Type.STOP, AudioSystem.NOT_SPECIFIED));
			}
		};
		stream.addUnderflowListener(stoppedListener);

		Stream.PlaybackStartedListener startedListener = new Stream.PlaybackStartedListener() {
			@Override
			public void update() {
				isEngagedInIo = true;
				fireLineEvent(new LineEvent(PulseAudioDataLine.this,
						LineEvent.Type.START, AudioSystem.NOT_SPECIFIED));
			}

		};

		stream.addPlaybackStartedListener(startedListener);

		try {
			synchronized (eventLoop.threadLock) {
				connectLine(bufferSize);
			}

		} catch (LineUnavailableException e) {
			// error connecting to the server!
			// FIXME clean up 
			throw e;
		}
		this.bufferSize = bufferSize;
		try {
			semaphore.acquire();
			synchronized (eventLoop.threadLock) {
				if (stream.getState() != Stream.State.READY) {
					stream.disconnect();
					throw new LineUnavailableException(
							"unable to obtain a line");
				}
			}
		} catch (InterruptedException e) {
			throw new LineUnavailableException("unable to prepare stream");
		}

	}

	public void open(AudioFormat format) throws LineUnavailableException {
		open(format, DEFAULT_BUFFER_SIZE);

	}

	public void open() throws LineUnavailableException {
		assert (defaultFormat != null);

		open(defaultFormat, DEFAULT_BUFFER_SIZE);
	}

	public void close() {

		synchronized (eventLoop.threadLock) {
			drain();
			stream.disconnect();
		}

		try {
			semaphore.acquire();
		} catch (InterruptedException e) {
			throw new RuntimeException("unable to prepare stream");
		}

		super.close();

	}

	public void start() {
		isStarted = true;
	}

	public void stop() {
		isStarted = false;
	}

	// A BIG FIXME !

	/*
	 * http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=4791152 :
	 * 
	 * a line is active in between calls to start() and stop(). In that sense,
	 * active means that the line is ready to take or give data. Running is
	 * tightly bound to data flow in the line. I.e. when you start a
	 * SourceDataLine but never write data to it, the line should not be
	 * running. This also means that a line should become not running on buffer
	 * underrun/overflow.
	 * 
	 * 
	 */

	public boolean isActive() {
		return isStarted;
	}

	public boolean isRunning() {
		return isEngagedInIo;
	}

	protected abstract void connectLine(int bufferSize)
			throws LineUnavailableException;

	public abstract void drain();

	public Stream getStream() {
		return stream;
	}

	@Override
	public int getBufferSize() {
		if (!isOpen) {
			return DEFAULT_BUFFER_SIZE;
		}
		return bufferSize;
	}

	public javax.sound.sampled.Line.Info getLineInfo() {
		return new DataLine.Info(this.getClass(), supportedFormats,
				StreamBufferAttributes.MIN_VALUE,
				StreamBufferAttributes.MAX_VALUE);
	}

	@Override
	public AudioFormat getFormat() {
		if (!isOpen) {
			return defaultFormat;
		}
		return currentFormat;
	}

	public float getLevel() {
		return AudioSystem.NOT_SPECIFIED;
	}

	public void setName(String streamName) {
		if (isOpen) {
			/*
			 * Note: setting the name of the stream after it's created wont
			 * work. In fact, it sets the name of the application! This is a bug
			 * in PulseAudio 0.9.12 but fixed in git.
			 */

			Operation o;
			synchronized (eventLoop.threadLock) {
				o = stream.setName(streamName);
			}
			o.waitForCompletion();
			o.releaseReference();

		}

		this.streamName = streamName;

	}

	public String getName() {
		return streamName;
	}

}
